/* GSI - general sound interface: NAS server connection code  [-*- C++ -*-]
 *
 * Copyright (C) W.H.Scholten 1997,1998
 *
 * gsi_nas_interface.c
 *
 * This interface file emulates the native sound driver
 * ** See gsi.h for a description of the functions **
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */

/* Note that this was based on SND_SERVER 0.47 and is being revised 
 * and updated for GSI */


/*---------------------------------------------------------------------------*
 * DEVEL INFO: Use the following NAS functions:
 *
 * AuSetCloseDownMode(server_name, close_mode, NULL);
 * (close_mode = AuCloseDownRetainTemporary/Permanent)
 * AuKillClient(server_name, resource, NULL);
 *
 * Another way to kill a sound:
 *    AuElementState states[1];
 *    AuMakeElementState(states, flow, AuElementAll, AuStateStop);
 *    AuSetElementStates(aud, 1, states, &status);
 *    AuHandleEvents(aud);
 *---------------------------------------------------------------------------*/


/*__________________INCLUDE___________________*/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>

#include <audio/audiolib.h>
#include <audio/soundlib.h>

#include <gsi/mytypes.h>
#include <gsi/debug.h>

#define  _GSI_INTERFACE_
#include <gsi/gsi.h>


/*__________________DEFINITIES___________________*/
#define	MAX_SOUNDS	256
#define MAX_QUEUE_LENGTH 256


/*__________________VARIABELEN___________________*/
char *gsi_version=GSI_INTERFACE_VERSION;

static int  sound_volume=80; //default PCM volume
static int  sound_on=1;

//internal:
char *sound_path=SOUNDDIR;

static char sndserver_ok = False;
static int  num_sounds = 0;

static int  nas_volume;

static int  num_commands;
static int  sound_command[MAX_QUEUE_LENGTH];

static char *SOUNDFILENAME[256];

static char  *server_name;
static AuServer    *aud = NULL;
static AuDeviceID  device;

static struct {
	char       *name;
	char       *filename;
	AuBucketID	bucket;
} sound_table[MAX_SOUNDS];


typedef struct {
	int         sound_playing_no; //index no. of the queue of playing sounds (for debugging only)
	int         sound_no; //sound table
	int	        playing;
	AuFlowID    flow;
} Audio_Record, *Audio_Ptr;


static Audio_Record playing_sounds[MAX_SOUNDS];

static int  num_playing;

/*__________________DECLARATIES__________________*/
static int  init_sounds_file(char *sounddir, char *sndsfile_name);
static void read_sounds();
static int  read_sound(int k);
static void kill_sound(int channel, int sound);

//specific NAS stuff
static void au_play_sound_file(int i, char *filename, int volume, int repeat);
static void au_kill_sound(AuFlowID flow);
static void au_check_sound();
static void audioDevicePlay(int i, char *filename, int volume, int repeat);
static void audioCallBack(AuServer *aud, AuEventHandlerRec *handler, AuEvent *event, AuPointer bucket_data);


/*__________________PROGRAMMA___________________*/

/*---------------------------------------------------------------------------*
 *	Main initialization, connects to the NAS server
 *---------------------------------------------------------------------------*/
int  gsi_init() {
	char *audio_status;
	int  i;

//  extern Display	*display;
//  server_name = XDisplayString(W_Display);
//  server_name=":0.0";
//  servername=getenv("DISPLAY");
	server_name="";

	DEB1(fprintf(stderr, "initialising NAS\n"));
	if (sndserver_ok)
		return 0; //already initialised
  
	/* try to connect to the NAS audio server */
	if (!(aud = AuOpenServer(server_name, 0, NULL, 0, NULL, NULL))) {
		fprintf(stderr, "%s: NAS connection failed\n", gsi_version);
		return -1;
	}

	/* Look for an audio device that we can use (mono) */
	for (i = 0; i < AuServerNumDevices(aud); i++) {
		if ((AuDeviceKind(AuServerDevice(aud, i)) == 
			 AuComponentKindPhysicalOutput) && 
			AuDeviceNumTracks(AuServerDevice(aud, i)) == 1) {
			device = AuDeviceIdentifier(AuServerDevice(aud, i));
			break;
		}
	}

	if (!device) {
		/* No device available - all busy or server connection failed? */
		AuCloseServer(aud);
		fprintf(stderr, "NAS: Connection to audio server failed\n");
		return -1;
	}
	sndserver_ok=True;
	AuSoundRestartHardwarePauses = AuFalse;

//   AuSetCloseDownMode(server_name, AuCloseDownRetainTemporary, NULL);


	DEB1(fprintf(stderr, "NAS sound initialised\n"));
	return 0;
}

/*---------------------------------------------------------------------------*
 *	Set sound volume
 *---------------------------------------------------------------------------*/
void set_volume(int vol) {
	DEB1(fprt(stderr, "NAS: set_volume to %d\n", vol));
	sound_volume=vol;
/*	nas_volume=volume_table[sound_volume];*/
	nas_volume=(vol*100)>>8;

	//Use NAS mixer functions (change sounds that are already playing)
	//i.e. use AuSetElementParameters for sounds that are running
}


#if 0
//unfinished new code: set volume of all playing sounds
static void adjust_volume(new_vol) {
	int             value = (int) new_vol;
	AuElementParameters *params;
	element_list_ptr  element = element_list;
	int             i = 0;

	if (!element_count) //list no, for the new list ==num_playing
		return;

	if (!(params = (AuElementParameters *) malloc(sizeof(AuElementParameters) * ElementCount))) {
		fprintf(stderr, "NAS: malloc error in adjust_volume");
		return;
	}

	while (element) {
		params[i].flow = element->flow;
		params[i].element_num = element->volume_element;
		params[i].num_parameters = AuParmsMultiplyConstant;
		params[i].parameters[AuParmsMultiplyConstantConstant] = AuFixedPointFromFraction(value, 100);
		element = element->next;
		i++;
	}

	AuSetElementParameters(aud, element_count, params, NULL);
	free(params);
}

//volume_element=multiplier from AuSoundPlayFromBucket( ..., &flow, &multiplier, ... )
#endif


/*---------------------------------------------------------------------------*
 *	Initialize PCM system
 *---------------------------------------------------------------------------*/
int  gsi_init_pcm(int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile) {
	if (init_sounds_file(sounddir, soundfile)==-1) sndserver_ok=False;
//   read_sounds();

	set_volume(sound_volume);

	/* Use environment variable for sound files instead of hard coding ? */
//      if ((string = getenv("LDEF_SOUND_DIR")) != NULL)
//	  sound_path=string;

	for (i=0; i<MAX_SOUNDS; i++) sound_table[i].name=NULL;

	return 0;
}


/*---------------------------------------------------------------------------*
 *	Init SYNTH system
 *---------------------------------------------------------------------------*/
int gsi_init_synth (int preferred_type) {
	fprintf(stderr, "NAS: gsi_init_synth is not yet implemented\n");
	return -1;
}


int gsi_close() {
	AuCloseServer(aud);
	aud=NULL;
	sndserver_ok=0;
//    AuKillClient(server_name, resource, NULL);

	return 0;
}


/*---------------------------------------------------------------------------*
 *	Play sounds 'n stuff
 *---------------------------------------------------------------------------*/
// Play a sound
void gsi_play_sound (int slot, int command) {
	DEB1(fprt(stderr, "NAS: play sound %d (command %d)\n", slot, command));

	if (sound_on && aud) {
		if (num_commands==MAX_QUEUE_LENGTH) gsi_flush();
		sound_command[num_commands++]=slot | (command<<24);
	}
}


void gsi_play_channel_sound (int channel, int slot, int command) {

}


void gsi_play_sound_3d (int ch, int slot, int command, int x, int y, int z, int dx, int dy, int dz, int doppler_on) {
}


// Play a sound with stereo position/movement
void gsi_play_sound_stereo (int ch, int slot, int command, int x, int dx, int doppler_on) {
	//effects are currently ignored
	ASSERT(ch>=-1 && ch <256);

	if (ch!=-1) gsi_play_sound (slot, command);
	else
		gsi_play_channel_sound (ch, slot, command);
}


void gsi_play_sound_pan (int ch, int slot, int command, int pan) {
	//effects are currently ignored
	ASSERT(ch>=-1 && ch <256);

	if (ch!=-1) gsi_play_sound (slot, command);
	else
		gsi_play_channel_sound (ch, slot, command);
}


void kill_sound(int channel, int sound) {
	int  i;

	DEB1(fprt(stderr, "NAS: kill sound %d, channel %d\n", sound, channel));

	if (channel==0)
		if (sound_table[sound].name !=NULL) {
			for (i=0; i<256; i++) {
				if (playing_sounds[i].sound_no==sound) {
					DEB1(fprt(stderr, "     -> sound found at queue no %d => playing=%d\n", i, playing_sounds[i].playing));
					if (playing_sounds[i].playing>=1) {
						playing_sounds[i].playing=-2; //indicating it will be killed
						au_kill_sound(playing_sounds[i].flow);
						DEB1(fprt(stderr, "     -> kill at queue no %d\n", i));
						break;
					}
				}
			}
		}
	DEB1(if (i==256) fprt(stderr, "     -> NOT FOUND\n"));
}


void gsi_kill_sounds() {
	int  i;

	num_commands=0;

	DEB1(fprt(stderr, "NAS: kill sounds\n"));

	for (i=0; i<num_sounds; i++) {
		kill_sound(0, i);
	}
	AuFlush(aud);
	AuHandleEvents(aud);
}


int gsi_play_song(int loopno) {
	ASSERT(loopno>=0 && loopno<256);
	gsi_send_commands(2, SND_PLAY_SONG, loopno);
	return 0;
}
int gsi_stop_song() {
	gsi_send_commands(1, SND_STOP_SONG);
	return 0;
}
int gsi_continue_song() {
	gsi_send_commands(1, SND_CONTINUE_SONG);
	return 0;
}

// execute the stored commands
// At the moment, sounds are ALWAYS assumed to want OVERWRITE.
void gsi_flush() {
	int i;
	byte command;
	int data, data2;

	DEB1(if (num_commands) fprt(stderr, "NAS: flush sounds (%d)\n", num_commands));
	for (i=0; i<num_commands; i++) {
		DEB1(fprt(stderr, "NAS: flush_sounds: command no %d\n", i));
		command=(sound_command[i]>>24)&0xff;
		data=sound_command[i]&0xff;
		data2=(sound_command[i]&0xff00)>>8;
		DEB1(fprt(stderr, "     (flush_sounds) command=%d, data=%d, data2=%d\n", command, data, data2));
		DEB1(fprt(stderr, "     (flush_sounds) file=%s\n", SOUNDFILENAME[data]));
		switch (command) {
		case SND_PLAY:
			DEB1(fprt(stderr, "     (flush_sounds) SND_PLAY\n"));
			au_play_sound_file(data, SOUNDFILENAME[data], nas_volume, 1);
			break;
		case SND_OVERWRITE:
			DEB1(fprt(stderr, "     (flush_sounds) SND_OVERWRITE\n"));
			au_play_sound_file(data, SOUNDFILENAME[data], nas_volume, 1);
			break;
		case SND_REPEAT:
			DEB1(fprt(stderr, "     (flush_sounds) SND_REPEAT\n"));
			au_play_sound_file(data, SOUNDFILENAME[data], nas_volume, 100);
			break;
		case SND_OVERWRITE_REPEAT:
			DEB1(fprt(stderr, "     (flush_sounds) SND_OVERWRITE_REPEAT\n"));
			au_play_sound_file(data, SOUNDFILENAME[data], nas_volume, 100);
			break;
		case SND_KILL:
			DEB1(fprt(stderr, "     (flush_sounds) SND_KILL\n"));
			kill_sound(0, data);
			DEB1(fprt(stderr, "     (flush_sounds) SND_KILL done, i=%d, num_commands=%d\n", i, num_commands));
			break;
		case SND_KILLALL:
			DEB1(fprt(stderr, "     (flush_sounds) SND_KILLALL\n"));
			kill_sounds();
			break;
		case SND_PAUSE:
		case SND_CONTINUE:
		case SND_GRAB_DSP:
		case SND_RELEASE_DSP:
			break;
		default: fprintf(stderr, "GSI on NAS: Bad sound command\n");
		}
	}
	num_commands=0;

	AuFlush(aud);
	AuHandleEvents(aud);
}


void gsi_play_sound_flush (int slot, int command) {
	gsi_play_sound (slot, command);
	gsi_flush();
}




/*---------------------------------------------------------------------------*
 *	Misc stuff
 *---------------------------------------------------------------------------*/
void gsi_grab_pcm() {
//gsi_send_commands(1, SND_GRAB_DSP);
	gsi_flush();
}


void gsi_release_pcm() {
//  gsi_send_commands(1, (SND_RELEASE_DSP);
	gsi_flush();
}



/*---------------------------------------------------------------------------*
 *	Init code from native sound server
 *---------------------------------------------------------------------------*/
int init_sounds_file(char *sounds_dir, char *sndsfile_name) {
	int  i;
	char s[1024]; /* conversion string, 128 should be enough */
	char s2[1024];
	FILE *fp;

	strcpy(s, sndsfile_name);

	fp = fopen(s, "r");
	if( fp == 0 ) {
		fprintf (stderr, "NAS: The sounds file %s could not be opened\n", sndsfile_name);
		return -1;
	}
	for (i=0; i<256; i++) {
		fgets(s2, 127, fp);
		if ( feof(fp) ) break;
		if ( strlen(s2)==0 ) break;
		s2[(int) strlen(s2) -1]=0; /* get rid of the linefeed */
		strcpy(s, sounds_dir);

		if ( s[(int)strlen(s) - 1] == '/' )
			strcat (s, s2+1);
		else
			strcat (s, s2);
//    SOUNDFILENAME[i] = (char *) malloc ((int)strlen (s)+1);
		SOUNDFILENAME[i] = new char[((int)strlen (s)+1)];

		strcpy (SOUNDFILENAME[i],s);
	}
	fclose(fp);
	num_sounds=i;

	DEB1(fprt(stderr, "NAS: sounds file loaded\n"));
	return 0;
}


/*---------------------------------------------------------------------------*
 *	Read all sound files from disk
 *  (to prevent hickups when a sound is first used)
 *---------------------------------------------------------------------------*/
void read_sounds() {
//    int i;

//  for (i=0; i<Num_Sounds; i++) {
//    read_sound(i);
//  }
}


/*---------------------------------------------------------------------------*
 *	Read a sound file from disk
 *---------------------------------------------------------------------------*/
int read_sound(int k) {
	// Not necessary, NAS does this (depends on NAS play command though)

#if 0 /* Not used at the moment */
	int i,fd,size, data_start;

	if ( k>=Num_Sounds ) {
		fprintf (stderr, "NAS %s: invalid sound number %d\n", server_name, k);
		return(0);
	};

	DEB1(fprt(stderr, "NAS: loading sound %d: %s",k,SOUNDFILENAME[k]));

	fd = open(SOUNDFILENAME[k], O_RDONLY);
	if( fd <= 0 )  {
		fprintf (stderr, "NAS %s: The sound %s could not be opened\n", server_name, SOUNDFILENAME[k]);
		sound_size[k]=-1;
		return(0);
	};
	// Check that the file is a regular file (not a dir)
	struct stat fd_stat;
	fstat(fd, &fd_stat);
	if (S_ISREG(fd_stat.st_mode)) {
		if (!strcmp(SOUNDFILENAME[k]+strlen(SOUNDFILENAME[k])-4,".raw")) {
			sound_speed[k]=dsp_server_speed;
			sound_type[k]=dsp_server_type;
			size=lseek(fd,0,SEEK_END);
			data_start=0;
		}
		else
			if (!strcmp(SOUNDFILENAME[k]+strlen(SOUNDFILENAME[k])-4,".au")) {
				lseek(fd,4,SEEK_SET);
				data_start=int_msb(fd);
				size=int_msb(fd);
				sound_type[k]=int_msb(fd);
				sound_speed[k]=int_msb(fd);
			}
			else {
				fprintf(stderr, "NAS %s: unrecognized file format of %s\n", server_name, SOUNDFILENAME[k]);
			}
		sound_size[k]=size;
//  sound_buffer[k]=(sbyte *) malloc(sound_size[k]);
		sound_buffer[k]=new sbyte[sound_size[k]];
		if(sound_buffer[k]==NULL) {
			fprintf(stderr,"NAS %s: couldn't malloc memory for sound\n", server_name);
			sound_size[k]=-1;
			close(fd);
			return(0);
		}
		lseek(fd,data_start,SEEK_SET);
		read(fd,sound_buffer[k],size);
	}
	else {
		DEB1(fprt(stderr, " %d not a regular file\n", k));
		close(fd); return 0;
	}
	close(fd);
	for(i=0;i<size;i++)  sound_buffer[k][i]^=0x80;

	DEB1(fprt(stderr, " -> loaded, %d bytes",size));
	DEB1(fprt(stderr, "read done\n"));
#endif /* Not used */
	return(1);
}


///////////////////////////// NAS SPECIFIC STUFF //////////////////////////////
/*---------------------------------------------------------------------------*
 *	NAS: check on a sound (see if it's running), UNFINISHED.
 *---------------------------------------------------------------------------*/
static void au_check_sound() {
	DEB1(fprt(stderr, "NAS: au_check_sound\n"));
	if (aud) AuHandleEvents(aud);
}


/*---------------------------------------------------------------------------*
 *	NAS: kill a sound, >>doesn't work nicely<<
 *---------------------------------------------------------------------------*/
static void au_kill_sound(AuFlowID flow) {
//   AuStatus status;

	DEB1(fprt(stderr, "NAS: au_kill_sound\n"));

	AuStopFlow(aud, flow, NULL);
//    AuStopFlow(aud, flow, &status);
//    AuFlush(aud);
//    AuHandleEvents(aud); //Need this for fast killing of the sound, but its SLOOOOOOOW.... (only when requesting status?)

}


/*---------------------------------------------------------------------------*
 *	NAS: actual command to play a sound file
 *---------------------------------------------------------------------------*/
static void au_play_sound_file(int i, char *filename, int volume, int repeat) {

	DEB1(fprt(stderr, "NAS: au_play_sound_file: %s\n", filename));

#if 1
	/* Load sound if not done yet */
	if (sound_table[i].name ==NULL) {
		sound_table[i].name = strdup(filename);
		DEB1(fprt(stderr, "     (au_play_sound_file) added file %s\n", filename));
		sound_table[i].filename = strdup(filename);
	}
#else
	/* Load sound if not done yet */
	for (i = 0; i < num_sounds; i++) {
		if (sound_table[i].name !=NULL)
			if (!strcmp(sound_table[i].name, filename)) {
				break; /* found */
			}
	}

	if (i == num_sounds) { //Not found
		/* add new sound to the table */
		for (i = 0; i < 256; i++) {
			if (sound_table[i].name ==NULL) {
				sound_table[i].name = strdup(filename);
				DEB1(fprt(stderr, "     (au_play_sound_file) added file %s\n", filename));
//	sprintf(fbuf, "%s/%s", sound_path, filename);
//      sound_table[num_sounds].filename = strdup(fbuf);
				sound_table[i].filename = strdup(filename);
				if (i>=num_sounds)
					num_sounds++;
				break;
			}
		}
	}
#endif
	audioDevicePlay(i, sound_table[i].filename, volume, repeat);
}


static void audioDevicePlay(int sound_table_no, char *filename, int volume, int repeat) {
	AuFlowID flow;
	int  i;

//    AuFlush(aud);
	AuHandleEvents(aud);

	DEB1(fprt(stderr, "NAS: audioDevicePlay\n"));
	DEB1(fprt(stderr, "     (audioDevicePlay) sound_no=%d (%s)\n", sound_table_no, filename));


	for (i=0; i<256; i++) {
		if (playing_sounds[i].playing>0 && playing_sounds[i].sound_no==sound_table_no) {
			DEB1(fprt(stderr, "    -> (audioDevicePlay) overwrite queue no %d \n", i));
			// Kill this sound (should only be done for overwrite)
			playing_sounds[i].playing=-2; //flag saying it has been killed (until the actual call back sets it to 0)
			au_kill_sound(playing_sounds[i].flow);
			break;
		}
	}

	for (i=0; i<256; i++) {
		if (playing_sounds[i].playing==0) {
			if (!sound_table[sound_table_no].bucket) {
				DEB1(fprt(stderr, "     (audioDevicePlay) creating bucket\n"));
				sound_table[sound_table_no].bucket = AuSoundCreateBucketFromFile(aud, filename, AuAccessAllMasks, NULL, NULL);
			}
			break;
		}
	}

	DEB1(fprt(stderr, "     (audioDevicePlay) playing sound %d (queue no %d)\n", sound_table_no, i));
	AuSoundPlayFromBucket(aud, sound_table[sound_table_no].bucket, /*device*/AuNone,
						  AuFixedPointFromFraction(volume, 100),
						  audioCallBack, (AuPointer) &playing_sounds[i],
						  repeat, &flow, NULL, NULL, NULL);
//    playing_sounds[i].playing=1;
	playing_sounds[i].playing=repeat;
	playing_sounds[i].flow=flow;
	playing_sounds[i].sound_no=sound_table_no;
	playing_sounds[i].sound_playing_no=i;

	AuFlush(aud);
	AuHandleEvents(aud);
}


/*---------------------------------------------------------------------------*
 *	NAS callback:
 *  NB. For repeating sounds there's a call back for each time the sound has
 *      been played (i.e. 5 call backs when loopno=5).
 *  Took a while before I found that out. Those damned NAS docs are useless.
 *---------------------------------------------------------------------------*/
static void audioCallBack(AuServer *aud, AuEventHandlerRec *handler, AuEvent *event, AuPointer audio_data) {
	Audio_Ptr   info = (Audio_Ptr) audio_data;

	DEB1(fprt(stderr, "NAS: audio call back\n"));
	DEB1(fprt(stderr, "     -> for sound %d (%s), playing=%d, queue no = %d\n", info->sound_no, sound_table[info->sound_no].filename, info->playing, info->sound_playing_no));

//    info->playing = 0;
	if (info->playing>0) info->playing--;
	if (info->playing<0) info->playing=0;

	DEB1(fprt(stderr, "     (audio callback) done\n"));
}

///////////////////////////// NAS SPECIFIC STUFF: END ////////////////////////

/*---------------------------------------------------------------------------*
 *	EOF gsi_nas_interface.c
 *---------------------------------------------------------------------------*/
